from typing import Dict, Union
from supervisely.annotation.tag import Tag
from supervisely.annotation.tag_meta import TagMeta, TagValueType
from supervisely.app.widgets import Widget
from supervisely.app.widgets import (
    Checkbox,
    Field,
    Empty,
    Input,
    InputNumber,
    RadioGroup,
)


class InputTag(Widget):
    def __init__(self, tag_meta: TagMeta, widget_id: int = None):
        self._tag_meta = tag_meta
        self._component = self._get_tag_component(tag_meta)

        super().__init__(widget_id=widget_id, file_path=__file__)

    def _get_tag_component(self, tag_meta: TagMeta):
        if tag_meta.value_type == str(TagValueType.NONE):
            return Checkbox(content=Field(content=Empty(), title=tag_meta.name))
        if tag_meta.value_type == str(TagValueType.ANY_NUMBER):
            return Checkbox(
                content=Field(
                    content=InputNumber(controls=False, debounce=500),
                    title=tag_meta.name,
                )
            )
        if tag_meta.value_type == str(TagValueType.ANY_STRING):
            return Checkbox(content=Field(content=Input(), title=tag_meta.name))
        if tag_meta.value_type == str(TagValueType.ONEOF_STRING):
            items = [
                RadioGroup.Item(pv, pv, Empty()) for pv in tag_meta.possible_values
            ]
            return Checkbox(
                content=Field(content=RadioGroup(items=items), title=tag_meta.name)
            )

    def get_tag_meta(self):
        return self._tag_meta

    def activate(self):
        tag_component = self._component
        tag_component.check()

    def deactivate(self):
        tag_component = self._component
        tag_component.uncheck()

    def is_active(self):
        return self._component.is_checked()

    @property
    def value(self):
        return self._get_value()

    @value.setter
    def value(self, value):
        self._set_value(value)

    def is_valid_value(self, value):
        return self._tag_meta.is_valid_value(value)
    
    def set(self, tag: Union[Tag, None]):
        if tag is None:
            self._set_default_value()
            self.deactivate()
        else:
            self._set_value(tag.value)
            self.activate()

    def get_tag(self):
        if not self.is_active():
            return None
        tag_value = self._get_value()
        return Tag(self._tag_meta, tag_value)

    def _get_value(self):
        tag_component = self._component
        content = tag_component._content._content
        if type(content) is Empty:
            return None
        else:
            return content.get_value()

    def _set_value(self, value):
        if not self.is_valid_value(value):
            raise ValueError(f'Tag value "{value}" is invalid')
        tag_component = self._component
        content = tag_component._content._content
        if type(content) is InputNumber:
            content.value = value
        if type(content) is Input:
            content.set_value(value)
        if type(content) is RadioGroup:
            content.set_value(value)

    def _set_default_value(self):
        tag_component = self._component
        content = tag_component._content._content
        if type(content) is InputNumber:
            content.value = 0
        if type(content) is Input:
            content.set_value("")
        if type(content) is RadioGroup:
            content.set_value(None)

    def get_json_data(self):
        return None

    def get_json_state(self) -> Dict:
        return None



import json
import os
from copy import deepcopy

from supervisely.imaging import image as sly_image
from supervisely.nn.inference.rest_constants import GET_OUTPUT_META, IMAGE, INFERENCE, MODEL, OUTPUT_META, \
    ANNOTATION, META, MODE, GPU_DEVICE

from supervisely.worker_api.interfaces import SingleImageInferenceInterface
from supervisely.project.project_meta import ProjectMeta
from supervisely.nn.hosted.deploy import ModelDeploy
from supervisely.function_wrapper import function_wrapper

from flask import Flask
from flask_restful import Resource, Api, reqparse



class RestInferenceServer:
    def __init__(self, model: SingleImageInferenceInterface, name, port=None):
        self._app = Flask(name)
        if port == '':
            port = None
        self._port = port

        api = Api(self._app)
        api.add_resource(RestInferenceServer.GetOutputMeta,
                         '/' + MODEL + '/' + GET_OUTPUT_META,
                         resource_class_kwargs={'model': model})
        api.add_resource(RestInferenceServer.Inference,
                         '/' + MODEL + '/' + INFERENCE,
                         resource_class_kwargs={'model': model})

    def run(self):
        self._app.run(debug=False, port=self._port, host='0.0.0.0')

    class GetOutputMeta(Resource):
        def __init__(self, model):
            self._model = model
            self._parser = reqparse.RequestParser()
            self._parser.add_argument(META)
            self._parser.add_argument(MODE)

        def post(self):
            args = self._parser.parse_args()

            meta = args[META]
            mode = args[MODE]

            data = {
                "request_type": GET_OUTPUT_META,
                "meta": json.loads(meta) if meta is not None else ProjectMeta().to_json(),
                "mode": json.loads(mode) if mode is not None else {}
            }

            response_json = self._model._final_processing(data)
            if 'out_meta' in response_json:
                return response_json['out_meta']
            if 'output_meta' in response_json:
                return response_json['output_meta']
            return response_json


    class Inference(Resource):
        def __init__(self, model):
            from werkzeug.datastructures import FileStorage
            self._model = model
            self._parser = reqparse.RequestParser()
            self._parser.add_argument(IMAGE, type=FileStorage, location='files', help="input image", required=True)
            self._parser.add_argument(ANNOTATION, location='files', type=FileStorage)
            self._parser.add_argument(META, location='files', type=FileStorage)
            self._parser.add_argument(MODE, location='files', type=FileStorage)

        def post(self):
            args = self._parser.parse_args()
            img_bytes = args[IMAGE].stream.read()
            img = sly_image.read_bytes(img_bytes)

            meta = args[META]
            ann = args[ANNOTATION]
            mode = args[MODE]

            data = {
                "request_type": INFERENCE,
                "meta": json.loads(meta.stream.read().decode("utf-8")) if meta is not None else ProjectMeta().to_json(),
                "annotation": json.loads(ann.stream.read().decode("utf-8")) if ann is not None else None,
                "mode": json.loads(mode.stream.read().decode("utf-8")) if mode is not None else {},
                'image_arr': img
            }
            return self._model._final_processing(data)


class ModelRest(ModelDeploy):
    def load_config(self):
        gpu_device = os.getenv(GPU_DEVICE, 0)
        self.config = deepcopy(ModelDeploy.config)
        self.config["model"]["gpu_device"] = gpu_device

    def run(self):
        raise RuntimeError("Method is not accessible")

    def _create_applier(self):
        model_applier = function_wrapper(self.model_applier_cls, task_model_config={})
        return model_applier



# coding: utf-8
from __future__ import annotations
import uuid
import json
from typing import Optional, Dict, List

from supervisely.project.project_meta import ProjectMeta
from supervisely._utils import take_with_default
from supervisely.api.module_api import ApiField
from supervisely.pointcloud_annotation.pointcloud_object_collection import (
    PointcloudObjectCollection,
)
from supervisely.video_annotation.constants import (
    FRAMES,
    DESCRIPTION,
    FRAMES_COUNT,
    TAGS,
    OBJECTS,
    KEY,
)
from supervisely.pointcloud_annotation.pointcloud_figure import PointcloudFigure
from supervisely.pointcloud_annotation.pointcloud_episode_frame_collection import (
    PointcloudEpisodeFrameCollection,
)
from supervisely.video_annotation.key_id_map import KeyIdMap
from supervisely.pointcloud_annotation.pointcloud_episode_tag_collection import (
    PointcloudEpisodeTagCollection,
)


class PointcloudEpisodeAnnotation:
    def __init__(
        self,
        frames_count: Optional[int] = None,
        objects: Optional[PointcloudObjectCollection] = None,
        frames: Optional[PointcloudEpisodeFrameCollection] = None,
        tags: Optional[PointcloudEpisodeTagCollection] = None,
        description: Optional[str] = "",
        key: uuid.UUID = None,
    ) -> None:
        self._frames_count = frames_count
        self._description = description
        self._frames = take_with_default(frames, PointcloudEpisodeFrameCollection())
        self._tags = take_with_default(tags, PointcloudEpisodeTagCollection())
        self._objects = take_with_default(objects, PointcloudObjectCollection())
        self._key = take_with_default(key, uuid.uuid4())

    def get_tags_on_frame(self, frame_index: int) -> PointcloudEpisodeTagCollection:
        frame = self._frames.get(frame_index, None)
        if frame is None:
            if frame_index < self.frames_count:
                return PointcloudEpisodeTagCollection([])
            else:
                raise ValueError(f"No frame with index {frame_index} in annotation.")
        tags = []
        for tag in self._tags:
            if frame_index >= tag.frame_range[0] and frame_index <= tag.frame_range[1]:
                tags.append(tag)
        return PointcloudEpisodeTagCollection(tags)

    def get_objects_on_frame(self, frame_index: int) -> PointcloudObjectCollection:
        frame = self._frames.get(frame_index, None)
        if frame is None:
            if frame_index < self.frames_count:
                return PointcloudObjectCollection([])
            else:
                raise ValueError(f"No frame with index {frame_index} in annotation.")
        frame_objects = {}
        for fig in frame.figures:
            if fig.parent_object.key() not in frame_objects.keys():
                frame_objects[fig.parent_object.key()] = fig.parent_object
        return PointcloudObjectCollection(list(frame_objects.values()))

    def get_figures_on_frame(self, frame_index: int) -> List[PointcloudFigure]:
        frame = self._frames.get(frame_index, None)
        if frame is None:
            if frame_index < self.frames_count:
                return PointcloudObjectCollection([])
            else:
                raise ValueError(f"No frame with index {frame_index} in annotation.")
        return frame.figures

    def to_json(self, key_id_map: KeyIdMap = None) -> Dict:
        res_json = {
            DESCRIPTION: self.description,
            KEY: self.key().hex,
            TAGS: self.tags.to_json(key_id_map),
            OBJECTS: self.objects.to_json(key_id_map),
            FRAMES_COUNT: self.frames_count,
            FRAMES: self.frames.to_json(key_id_map),
        }

        if key_id_map is not None:
            dataset_id = key_id_map.get_video_id(self.key())
            if dataset_id is not None:
                res_json[ApiField.DATASET_ID] = dataset_id

        return res_json

    @classmethod
    def from_json(
        cls, data: Dict, project_meta: ProjectMeta, key_id_map: Optional[KeyIdMap] = None
    ):
        item_key = uuid.UUID(data[KEY]) if KEY in data else uuid.uuid4()

        if key_id_map is not None:
            key_id_map.add_video(item_key, data.get(ApiField.DATASET_ID, None))

        description = data.get(DESCRIPTION, "")
        frames_count = data.get(FRAMES_COUNT, 0)

        tags = PointcloudEpisodeTagCollection.from_json(
            data[TAGS], project_meta.tag_metas, key_id_map
        )
        objects = PointcloudObjectCollection.from_json(data[OBJECTS], project_meta, key_id_map)
        frames = PointcloudEpisodeFrameCollection.from_json(
            data[FRAMES], objects, key_id_map=key_id_map
        )

        return cls(frames_count, objects, frames, tags, description, item_key)

    @classmethod
    def load_json_file(
        cls, path: str, project_meta: ProjectMeta, key_id_map: Optional[KeyIdMap] = None
    ) -> PointcloudEpisodeAnnotation:
        """
        Loads json file and converts it to PointcloudEpisodeAnnotation.

        :param path: Path to the json file.
        :type path: str
        :param project_meta: Input :class:`ProjectMeta<supervisely.project.project_meta.ProjectMeta>`.
        :type project_meta: ProjectMeta
        :param key_id_map: KeyIdMap object.
        :type key_id_map: KeyIdMap, optional
        :return: PointcloudEpisodeAnnotation object
        :rtype: :class:`PointcloudEpisodeAnnotation<PointcloudEpisodeAnnotation>`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            address = 'https://app.supervise.ly/'
            token = 'Your Supervisely API Token'
            api = sly.Api(address, token)

            team_name = 'Vehicle Detection'
            workspace_name = 'Cities'
            project_name =  'London'

            team = api.team.get_info_by_name(team_name)
            workspace = api.workspace.get_info_by_name(team.id, workspace_name)
            project = api.project.get_info_by_name(workspace.id, project_name)

            meta = api.project.get_meta(project.id)

            # Load json file
            path = "/home/admin/work/docs/my_dataset/ann/annotation.json"
            ann = sly.PointcloudEpisodeAnnotation.load_json_file(path, meta)
        """
        with open(path) as fin:
            data = json.load(fin)
        return cls.from_json(data, project_meta, key_id_map)

    def clone(
        self,
        frames_count: Optional[int] = None,
        objects: Optional[PointcloudObjectCollection] = None,
        frames: Optional[PointcloudEpisodeFrameCollection] = None,
        tags: Optional[PointcloudEpisodeTagCollection] = None,
        description: Optional[str] = "",
    ) -> PointcloudEpisodeAnnotation:
        return PointcloudEpisodeAnnotation(
            frames_count=take_with_default(frames_count, self.frames_count),
            objects=take_with_default(objects, self.objects),
            frames=take_with_default(frames, self.frames),
            tags=take_with_default(tags, self.tags),
            description=take_with_default(description, self.description),
        )

    @property
    def frames_count(self) -> int:
        return self._frames_count

    @property
    def objects(self) -> PointcloudObjectCollection:
        return self._objects

    @property
    def frames(self) -> PointcloudEpisodeFrameCollection:
        return self._frames

    @property
    def figures(self) -> List[PointcloudFigure]:
        return self.frames.figures

    @property
    def tags(self) -> PointcloudEpisodeTagCollection:
        return self._tags

    def key(self) -> uuid.UUID:
        return self._key

    @property
    def description(self) -> str:
        return self._description

    def is_empty(self) -> bool:
        if len(self.objects) == 0 and len(self.tags) == 0:
            return True
        else:
            return False



from supervisely.app import DataJson, StateJson
from supervisely.app.widgets import Widget

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


class Input(Widget):
    def __init__(
            self,
            value: str = "",
            minlength: int = 0,
            maxlength: int = 1000,
            placeholder: str = "",
            size: Literal["mini", "small", "large"] = None,
            readonly: bool = False,
            widget_id: str = None
    ):
        self._value = value
        self._minlength = minlength
        self._maxlength = maxlength
        self._placeholder = placeholder
        self._size = size
        self._readonly = readonly
        self._widget_id = widget_id

        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self):
        return {
            "minlength": self._minlength,
            "maxlength": self._maxlength,
            "placeholder": self._placeholder,
            "size": self._size,
            "readonly": self._readonly,
        }

    def get_json_state(self):
        return {"value": self._value}

    def set_value(self, value):
        StateJson()[self.widget_id]["value"] = value
        StateJson().send_changes()

    def get_value(self):
        return StateJson()[self.widget_id]["value"]

    def is_readonly(self):
        return DataJson()[self.widget_id]["readonly"]

    def enable_readonly(self):
        DataJson()[self.widget_id]["readonly"] = True
        DataJson().send_changes()

    def disable_readonly(self):
        DataJson()[self.widget_id]["readonly"] = False
        DataJson().send_changes()



import copy
import uuid
import time

import supervisely
from supervisely.app import DataJson
from supervisely.app.widgets import Widget


class GridGallery(Widget):
    class Routes:
        IMAGE_CLICKED = "image_clicked_cb"

    def __init__(
        self,
        columns_number: int,
        annotations_opacity: float = 0.5,
        show_opacity_slider: bool = True,
        enable_zoom: bool = False,
        resize_on_zoom: bool = False,
        sync_views: bool = False,
        fill_rectangle: bool = True,
        border_width: int = 3,
        widget_id: str = None,
    ):

        self._data = []
        self._layout = []
        self._annotations = {}

        self.columns_number = columns_number

        self._last_used_column_index = 0
        self._project_meta: supervisely.ProjectMeta = None
        self._loading = False

        #############################
        # grid gallery settings
        self._show_preview: bool = True
        self._fill_rectangle: bool = fill_rectangle
        self._border_width: int = border_width

        self._show_opacity_header: bool = show_opacity_slider
        self._opacity: float = annotations_opacity
        self._enable_zoom: bool = enable_zoom
        self._sync_views: bool = sync_views
        self._resize_on_zoom: bool = resize_on_zoom
        #############################

        super().__init__(widget_id=widget_id, file_path=__file__)

    def _generate_project_meta(self):
        objects_dict = dict()

        for cell_data in self._data:
            annotation: supervisely.Annotation = cell_data["annotation"]
            for label in annotation.labels:
                objects_dict[label.obj_class.name] = label.obj_class

        objects_list = list(objects_dict.values())
        objects_collection = (
            supervisely.ObjClassCollection(objects_list) if len(objects_list) > 0 else None
        )

        self._project_meta = supervisely.ProjectMeta(obj_classes=objects_collection)
        return self._project_meta.to_json()

    def get_json_data(self):
        return {
            "content": {
                "projectMeta": self._generate_project_meta(),
                "layout": self._layout,
                "annotations": self._annotations,
            },
            "loading": self._loading,
        }

    def get_json_state(self):
        return {
            "options": {
                "showOpacityInHeader": self._show_opacity_header,
                "opacity": self._opacity,
                "enableZoom": self._enable_zoom,
                "syncViews": self._sync_views,
                "resizeOnZoom": self._resize_on_zoom,
                "fillRectangle": self._fill_rectangle,
                "borderWidth": self._border_width,
                "selectable": False,
                "viewHeight": None,
            },
            "selectedImage": None,
            "activeFigure": None,
        }

    def get_column_index(self, incoming_value):
        if incoming_value is not None and 0 > incoming_value > self.columns_number:
            raise ValueError(f"column index == {incoming_value} is out of bounds")

        if incoming_value is None:
            incoming_value = self._last_used_column_index
            self._last_used_column_index = (self._last_used_column_index + 1) % self.columns_number
        else:
            self._last_used_column_index = incoming_value

        return incoming_value

    def append(
        self,
        image_url: str,
        annotation: supervisely.Annotation = None,
        title: str = "",
        column_index: int = None,
        zoom_to: int = None,
        zoom_factor: float = 1.2,
    ):

        column_index = self.get_column_index(incoming_value=column_index)
        cell_uuid = str(
            uuid.uuid5(namespace=uuid.NAMESPACE_URL, name=f"{image_url}_{title}_{column_index}_{time.time()}").hex
        )

        self._data.append(
            {
                "image_url": image_url,
                "annotation": supervisely.Annotation((1, 1))
                if annotation is None
                else annotation.clone(),
                "column_index": column_index,
                "title": title,
                "cell_uuid": cell_uuid,
                "zoom_to": zoom_to,
                "zoom_factor": zoom_factor,
            }
        )

        self._update()

    def clean_up(self):
        self._data = []
        self._layout = []
        self._annotations = {}
        self._update()
        self.update_data()

    def _update_layout(self):
        layout = [[] for _ in range(self.columns_number)]

        for cell_data in self._data:
            layout[cell_data["column_index"]].append(cell_data["cell_uuid"])

        self._layout = copy.deepcopy(layout)
        DataJson()[self.widget_id]["content"]["layout"] = self._layout

    def _update_annotations(self):
        annotations = {}

        for cell_data in self._data:
            annotations[cell_data["cell_uuid"]] = {
                "url": cell_data["image_url"],
                "figures": [label.to_json() for label in cell_data["annotation"].labels],
                "title": cell_data["title"],
            }
            if not cell_data["zoom_to"] is None: 
                zoom_params = {
                    "figureId": cell_data["zoom_to"],
                    "factor": cell_data["zoom_factor"],
                }
                annotations[cell_data["cell_uuid"]]["zoomToFigure"] = zoom_params

        self._annotations = copy.deepcopy(annotations)
        DataJson()[self.widget_id]["content"]["annotations"] = self._annotations

    def _update_project_meta(self):
        DataJson()[self.widget_id]["content"]["projectMeta"] = self._generate_project_meta()

    def _update(self):
        self._update_layout()
        self._update_annotations()
        self._update_project_meta()

    @property
    def loading(self):
        return self._loading

    @loading.setter
    def loading(self, value: bool):
        self._loading = value
        DataJson()[self.widget_id]["loading"] = self._loading
        DataJson().send_changes()



# coding: utf-8
import uuid
from supervisely.video_annotation.video_figure import VideoFigure
from supervisely.video_annotation.key_id_map import KeyIdMap
from supervisely.geometry.closed_surface_mesh import ClosedSurfaceMesh
from supervisely.api.module_api import ApiField
from supervisely.geometry.any_geometry import AnyGeometry
from supervisely.annotation.json_geometries_map import GET_GEOMETRY_FROM_STR
from supervisely._utils import take_with_default
from supervisely.volume_annotation.volume_object import VolumeObject
import supervisely.volume_annotation.constants as constants
from supervisely.volume_annotation.constants import ID, KEY, OBJECT_ID, OBJECT_KEY, META
from supervisely.geometry.constants import (
    LABELER_LOGIN,
    UPDATED_AT,
    CREATED_AT,
    CLASS_ID,
)

from supervisely.volume_annotation.volume_object_collection import (
    VolumeObjectCollection,
)


class VolumeFigure(VideoFigure):
    def __init__(
        self,
        volume_object,
        geometry,
        plane_name,
        slice_index,
        key=None,
        class_id=None,
        labeler_login=None,
        updated_at=None,
        created_at=None,
    ):
        super().__init__(
            video_object=volume_object,
            geometry=geometry,
            frame_index=slice_index,
            key=key,
            class_id=class_id,
            labeler_login=labeler_login,
            updated_at=updated_at,
            created_at=created_at,
        )
        from supervisely.volume_annotation.plane import Plane

        Plane.validate_name(plane_name)
        self._plane_name = plane_name
        self._slice_index = slice_index

    @property
    def volume_object(self) -> VolumeObject:
        return self._video_object

    @property
    def video_object(self):
        raise NotImplementedError(
            'Property "video_object" is only available for videos'
        )

    @property
    def parent_object(self) -> VolumeObject:
        return self.volume_object

    @property
    def frame_index(self):
        raise NotImplementedError('Property "frame_index" is only available for videos')

    @property
    def slice_index(self):
        return self._slice_index

    @property
    def plane_name(self):
        return self._plane_name

    @property
    def normal(self):
        from supervisely.volume_annotation.plane import Plane

        return Plane.get_normal(self.plane_name)

    def _validate_geometry_type(self):
        if (
            self.parent_object.obj_class.geometry_type != AnyGeometry
            and type(self._geometry) != ClosedSurfaceMesh
        ):
            if type(self._geometry) is not self.parent_object.obj_class.geometry_type:
                raise RuntimeError(
                    "Input geometry type {!r} != geometry type of ObjClass {}".format(
                        type(self._geometry), self.parent_object.obj_class.geometry_type
                    )
                )

    def _validate_geometry(self):
        if type(self._geometry) == ClosedSurfaceMesh:
            return
        super()._validate_geometry()

    def validate_bounds(self, img_size, _auto_correct=False):
        if type(self._geometry) == ClosedSurfaceMesh:
            return
        super().validate_bounds(img_size, _auto_correct)

    def clone(
        self,
        volume_object=None,
        geometry=None,
        plane_name=None,
        slice_index=None,
        key=None,
        class_id=None,
        labeler_login=None,
        updated_at=None,
        created_at=None,
    ):
        return self.__class__(
            volume_object=take_with_default(volume_object, self.parent_object),
            geometry=take_with_default(geometry, self.geometry),
            plane_name=take_with_default(plane_name, self.plane_name),
            slice_index=take_with_default(slice_index, self.slice_index),
            key=take_with_default(key, self._key),
            class_id=take_with_default(class_id, self.class_id),
            labeler_login=take_with_default(labeler_login, self.labeler_login),
            updated_at=take_with_default(updated_at, self.updated_at),
            created_at=take_with_default(created_at, self.created_at),
        )

    def get_meta(self):
        return {
            constants.SLICE_INDEX: self.slice_index,
            constants.PLANE_NAME: self.plane_name,
            constants.NORMAL: self.normal,
        }

    @classmethod
    def from_json(
        cls,
        data,
        objects: VolumeObjectCollection,
        plane_name,
        slice_index,
        key_id_map: KeyIdMap = None,
    ):
        # @#TODO: copypaste from video figure, add base class and refactor copypaste later
        # _video_figure = super().from_json(data, objects, slice_index, key_id_map)

        object_id = data.get(OBJECT_ID, None)
        object_key = None
        if OBJECT_KEY in data:
            object_key = uuid.UUID(data[OBJECT_KEY])

        if object_id is None and object_key is None:
            raise RuntimeError(
                "Figure can not be deserialized from json: object_id or object_key are not found"
            )

        if object_key is None:
            if key_id_map is None:
                raise RuntimeError("Figure can not be deserialized: key_id_map is None")
            object_key = key_id_map.get_object_key(object_id)
            if object_key is None:
                raise RuntimeError(
                    "Object with id={!r} not found in key_id_map".format(object_id)
                )

        volume_object = objects.get(object_key)
        if volume_object is None:
            raise RuntimeError(
                "Figure can not be deserialized: corresponding object {!r} not found in ObjectsCollection".format(
                    object_key.hex
                )
            )

        shape_str = data[ApiField.GEOMETRY_TYPE]
        shape = GET_GEOMETRY_FROM_STR(shape_str)
        if shape == ClosedSurfaceMesh:
            geometry_json = data
        else:
            geometry_json = data[ApiField.GEOMETRY]
        geometry = shape.from_json(geometry_json)

        key = uuid.UUID(data[KEY]) if KEY in data else uuid.uuid4()

        if key_id_map is not None:
            key_id_map.add_figure(key, data.get(ID, None))

        class_id = data.get(CLASS_ID, None)
        labeler_login = data.get(LABELER_LOGIN, None)
        updated_at = data.get(UPDATED_AT, None)
        created_at = data.get(CREATED_AT, None)

        return cls(
            volume_object=volume_object,
            geometry=geometry,
            plane_name=plane_name,
            slice_index=slice_index,
            key=key,
            class_id=class_id,
            labeler_login=labeler_login,
            updated_at=updated_at,
            created_at=created_at,
        )

    def to_json(self, key_id_map=None, save_meta=True):
        json_data = super().to_json(key_id_map, save_meta)
        if type(self._geometry) == ClosedSurfaceMesh:
            json_data.pop(ApiField.GEOMETRY)
            json_data.pop(ApiField.META)
        return json_data



# coding: utf-8

from collections import namedtuple
from typing import Optional, List
import os

from supervisely.io.fs import file_exists, touch
from supervisely.io.json import dump_json_file, load_json_file
from supervisely.project.project_meta import ProjectMeta
from supervisely.task.progress import Progress
from supervisely._utils import batched
from supervisely.video_annotation.key_id_map import KeyIdMap

from supervisely.api.api import Api
from supervisely.api.module_api import ApiField
from supervisely.collection.key_indexed_collection import KeyIndexedCollection
from supervisely.volume import volume as sly_volume

from supervisely.project.project import Dataset, Project, OpenMode
from supervisely.project.video_project import VideoDataset, VideoProject
from supervisely.project.project import read_single_project as read_project_wrapper
from supervisely.project.project_type import ProjectType
from supervisely.volume_annotation.volume_annotation import VolumeAnnotation

VolumeItemPaths = namedtuple("VolumeItemPaths", ["volume_path", "ann_path"])


class VolumeDataset(VideoDataset):
    item_dir_name = "volume"
    interpolation_dir = "interpolation"
    annotation_class = VolumeAnnotation
    item_module = sly_volume
    paths_tuple = VolumeItemPaths

    @classmethod
    def _has_valid_ext(cls, path: str) -> bool:
        """
        Checks if file from given path is supported
        :param path: str
        :return: bool
        """
        return sly_volume.has_valid_ext(path)

    def _get_empty_annotaion(self, item_name):
        path = item_name
        _, volume_meta = sly_volume.read_nrrd_serie_volume(path)
        return self.annotation_class(volume_meta)

    def get_interpolation_dir(self, item_name):
        return os.path.join(self.directory, self.interpolation_dir, item_name)

    def get_interpolation_path(self, item_name, figure):
        return os.path.join(
            self.get_interpolation_dir(item_name), figure.key().hex + ".stl"
        )

    def get_classes_stats(
        self,
        project_meta: Optional[ProjectMeta] = None,
        return_objects_count: Optional[bool] = True,
        return_figures_count: Optional[bool] = True,
        return_items_count: Optional[bool] = True,
    ):
        if project_meta is None:
            project = VolumeProject(self.project_dir, OpenMode.READ)
            project_meta = project.meta
        class_items = {}
        class_objects = {}
        class_figures = {}
        for obj_class in project_meta.obj_classes:
            class_items[obj_class.name] = 0
            class_objects[obj_class.name] = 0
            class_figures[obj_class.name] = 0
        for item_name in self:
            item_ann = self.get_ann(item_name, project_meta)
            item_class = {}
            for ann_obj in item_ann.objects:
                class_objects[ann_obj.obj_class.name] += 1
            for volume_figure in item_ann.figures:
                class_figures[volume_figure.parent_object.obj_class.name] += 1
                item_class[volume_figure.parent_object.obj_class.name] = True
            for obj_class in project_meta.obj_classes:
                if obj_class.name in item_class.keys():
                    class_items[obj_class.name] += 1
        
        result = {}
        if return_items_count:
            result["items_count"] = class_items
        if return_objects_count:
            result["objects_count"] = class_objects
        if return_figures_count:
            result["figures_count"] = class_figures
        return result


class VolumeProject(VideoProject):
    dataset_class = VolumeDataset

    class DatasetDict(KeyIndexedCollection):
        item_type = VolumeDataset

    def get_classes_stats(
        self,
        dataset_names: Optional[List[str]] = None,
        return_objects_count: Optional[bool] = True,
        return_figures_count: Optional[bool] = True,
        return_items_count: Optional[bool] = True,
    ):
        return super(VolumeProject, self).get_classes_stats(
            dataset_names, 
            return_objects_count, 
            return_figures_count, 
            return_items_count
        )


def download_volume_project(
    api: Api,
    project_id,
    dest_dir,
    dataset_ids=None,
    download_volumes=True,
    log_progress=False,
):
    LOG_BATCH_SIZE = 1

    key_id_map = KeyIdMap()

    project_fs = VolumeProject(dest_dir, OpenMode.CREATE)

    meta = ProjectMeta.from_json(api.project.get_meta(project_id))
    project_fs.set_meta(meta)

    datasets_infos = []
    if dataset_ids is not None:
        for ds_id in dataset_ids:
            datasets_infos.append(api.dataset.get_info_by_id(ds_id))
    else:
        datasets_infos = api.dataset.get_list(project_id)

    for dataset in datasets_infos:
        dataset_fs = project_fs.create_dataset(dataset.name)
        volumes = api.volume.get_list(dataset.id)

        ds_progress = None
        if log_progress:
            ds_progress = Progress(
                "Downloading dataset: {!r}".format(dataset.name), total_cnt=len(volumes)
            )
        for batch in batched(volumes, batch_size=LOG_BATCH_SIZE):
            volume_ids = [volume_info.id for volume_info in batch]
            volume_names = [volume_info.name for volume_info in batch]

            ann_jsons = api.volume.annotation.download_bulk(dataset.id, volume_ids)

            for volume_id, volume_name, volume_info, ann_json in zip(
                volume_ids, volume_names, batch, ann_jsons
            ):
                if volume_name != ann_json[ApiField.VOLUME_NAME]:
                    raise RuntimeError(
                        "Error in api.volume.annotation.download_batch: broken order"
                    )

                volume_file_path = dataset_fs.generate_item_path(volume_name)
                if download_volumes is True:
                    item_progress = None
                    if log_progress:
                        item_progress = Progress(
                            f"Downloading {volume_name}",
                            total_cnt=volume_info.sizeb,
                            is_size=True,
                        )
                    api.volume.download_path(
                        volume_id, volume_file_path, item_progress.iters_done_report
                    )
                else:
                    touch(volume_file_path)

                ann = VolumeAnnotation.from_json(ann_json, project_fs.meta, key_id_map)
                dataset_fs.add_item_file(
                    volume_name,
                    volume_file_path,
                    ann=ann,
                    _validate_item=False,
                )

                mesh_ids = []
                mesh_paths = []
                for sf in ann.spatial_figures:
                    figure_id = key_id_map.get_figure_id(sf.key())
                    mesh_ids.append(figure_id)
                    figure_path = dataset_fs.get_interpolation_path(volume_name, sf)
                    mesh_paths.append(figure_path)
                api.volume.figure.download_stl_meshes(mesh_ids, mesh_paths)

            ds_progress.iters_done_report(len(batch))
    project_fs.set_key_id_map(key_id_map)


# TODO: add methods to convert to 3d masks


def upload_volume_project(
    dir, api: Api, workspace_id, project_name=None, log_progress=True
):
    project_fs = VolumeProject.read_single(dir)
    if project_name is None:
        project_name = project_fs.name

    if api.project.exists(workspace_id, project_name):
        project_name = api.project.get_free_name(workspace_id, project_name)

    project = api.project.create(workspace_id, project_name, ProjectType.VOLUMES)
    api.project.update_meta(project.id, project_fs.meta.to_json())

    for dataset_fs in project_fs.datasets:
        dataset = api.dataset.create(project.id, dataset_fs.name)

        names, item_paths, ann_paths, interpolation_dirs = [], [], [], []
        for item_name in dataset_fs:
            img_path, ann_path = dataset_fs.get_item_paths(item_name)
            names.append(item_name)
            item_paths.append(img_path)
            ann_paths.append(ann_path)
            interpolation_dirs.append(dataset_fs.get_interpolation_dir(item_name))

        progress_cb = None
        if log_progress:
            ds_progress = Progress(
                "Uploading volumes to dataset {!r}".format(dataset.name),
                total_cnt=len(item_paths),
            )
            progress_cb = ds_progress.iters_done_report

        item_infos = api.volume.upload_nrrd_series_paths(
            dataset.id, names, item_paths, progress_cb
        )
        item_ids = [item_info.id for item_info in item_infos]
        ds_progress = None
        if log_progress:
            ds_progress = Progress(
                "Uploading annotations to dataset {!r}".format(dataset.name),
                total_cnt=len(item_paths),
            )
            progress_cb = ds_progress.iters_done_report

        api.volume.annotation.upload_paths(
            item_ids, ann_paths, project_fs.meta, interpolation_dirs, progress_cb
        )

    return project.id, project.name



from typing import List

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal
from supervisely.app import DataJson
from supervisely.app.widgets import Widget


class Container(Widget):
    def __init__(
        self,
        widgets: List[Widget] = [],
        direction: Literal["vertical", "horizontal"] = "vertical",
        gap: int = 10,
        fractions: List[int] = None,
        # overflow: Literal["scroll", "wrap"] = None,
        # grid_cell_width: Literal["20%", "300px"] = None,
        widget_id: str = None,
    ):
        self._widgets = widgets
        self._direction = direction
        self._gap = gap
        self._overflow = "scroll"
        self._grid_cell_width = None  # grid_cell_width
        if self._direction == "vertical" and fractions is not None:
            raise ValueError("fractions can be defined only with horizontal direction")

        if fractions is not None and len(widgets) != len(fractions):
            raise ValueError(
                "len(widgets) != len(fractions): fractions have to be defined for all widgets"
            )

        if self._direction == "vertical":
            self._overflow = None
            if self._overflow is not None:
                raise ValueError(
                    "overflow argument can only be defined if direction is 'horizontal'"
                )
            if self._grid_cell_width is not None:
                raise ValueError(
                    "grid_cell_width argument can only be defined if direction is 'horizontal'"
                )

        if self._direction == "horizontal" and self._overflow is None:
            self._overflow = "wrap"

        if self._grid_cell_width is not None and self._overflow != "wrap":
            raise ValueError(
                "grid_cell_width argument can only be defined if overflow is 'wrap'"
            )

        self._fractions = fractions
        self._flex_direction = "column"
        if direction == "horizontal":
            self._flex_direction = "row"
            if self._fractions is None:
                if self._grid_cell_width is None:
                    self._fractions = ["1 1 auto"] * len(self._widgets)
                else:
                    self._fractions = [
                        f"1 1 calc({self._grid_cell_width} - {self._gap}px)"
                    ] * len(self._widgets)
                # self._fractions = ["1"] * len(self._widgets)
        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self):
        return None

    def get_json_state(self):
        return None



import typing
import os
from os import PathLike
import jinja2
from fastapi.templating import Jinja2Templates as _fastapi_Jinja2Templates
from starlette.templating import _TemplateResponse as _TemplateResponse
from starlette.background import BackgroundTask
from supervisely.app.singleton import Singleton
from supervisely.app.widgets_context import JinjaWidgets

# https://github.com/supervisely/js-bundle
js_bundle_version = "2.1.30"

# https://github.com/supervisely-ecosystem/supervisely-app-frontend-js
js_frontend_version = "0.0.35"


class Jinja2Templates(_fastapi_Jinja2Templates, metaclass=Singleton):
    def __init__(self, directory: typing.Union[str, PathLike] = "templates") -> None:
        super().__init__(directory)

    def _create_env(self, directory: typing.Union[str, PathLike]) -> "jinja2.Environment":
        env_fastapi = super()._create_env(directory)
        env_sly = jinja2.Environment(
            loader=env_fastapi.loader,
            autoescape=True,
            variable_start_string="{{{",
            variable_end_string="}}}",
        )
        env_sly.globals["url_for"] = env_fastapi.globals["url_for"]
        return env_sly

    def TemplateResponse(
        self,
        name: str,
        context: dict,
        status_code: int = 200,
        headers: dict = None,
        media_type: str = None,
        background: BackgroundTask = None,
    ) -> _TemplateResponse:
        from supervisely.app.fastapi.subapp import get_name_from_env

        context_with_widgets = {
            **context,
            **JinjaWidgets().context,
            "js_bundle_version": js_bundle_version,
            "js_frontend_version": js_frontend_version,
            "app_name": get_name_from_env(default="Supervisely App"),
        }

        return super().TemplateResponse(
            name, context_with_widgets, status_code, headers, media_type, background
        )



# coding: utf-8
"""api for working with agent"""

from __future__ import annotations

from typing import NamedTuple, Optional, Dict, List
from enum import Enum
from supervisely.api.module_api import ApiField, ModuleApi, ModuleWithStatus


class AgentNotFound(Exception):
    """
    """
    pass


class AgentNotRunning(Exception):
    """
    """
    pass


class AgentApi(ModuleApi, ModuleWithStatus):
    """
    API for working with agent. :class:`AgentApi<AgentApi>` object is immutable.

    :param api: API connection to the server
    :type api: Api
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        # You can connect to API directly
        address = 'https://app.supervise.ly/'
        token = 'Your Supervisely API Token'
        api = sly.Api(address, token)

        # Or you can use API from environment
        os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
        os.environ['API_TOKEN'] = 'Your Supervisely API Token'
        api = sly.Api.from_env()

        team_id = 8
        agents = api.agent.get_list(team_id)
    """

    class Status(Enum):
        """Agent status."""

        WAITING = 'waiting'
        """"""
        RUNNING = 'running'
        """"""

    @staticmethod
    def info_sequence():
        """
        NamedTuple AgentInfo information about Agent.

        :Example:

         .. code-block:: python

            AgentInfo("some info")
        """
        return [ApiField.ID,
                ApiField.NAME,
                ApiField.TOKEN,
                ApiField.STATUS,
                ApiField.USER_ID,
                ApiField.TEAM_ID,
                ApiField.CAPABILITIES,
                ApiField.CREATED_AT,
                ApiField.UPDATED_AT]

    @staticmethod
    def info_tuple_name():
        """
        NamedTuple name - **AgentInfo**.
        """
        return 'AgentInfo'

    def __init__(self, api):
        ModuleApi.__init__(self, api)
        ModuleWithStatus.__init__(self)

    def get_list(self, team_id: int, filters: Optional[List[Dict[str, str]]] = None) -> List[NamedTuple]:
        """
        List of all agents in the given Team.

        :param team_id: Team ID in Supervisely.
        :type team_id: int
        :param filters: List of params to sort output Agents.
        :type filters: List[dict], optional
        :return: List of Agents with information. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`List[NamedTuple]`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            team_id = 16087
            agents = api.agent.get_list(team_id)

            filter_agents = api.agent.get_list(team_id, filters=[{ 'field': 'name', 'operator': '=', 'value': 'Gorgeous Chicken' }])
        """
        return self.get_list_all_pages('agents.list', {'teamId': team_id, "filter": filters or []})

    def get_info_by_id(self, id: int) -> NamedTuple:
        """
        Get Agent information by ID.

        :param id: Agent ID in Supervisely.
        :type id: int
        :return: Information about Agent. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`NamedTuple`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            agent = api.agent.get_info_by_id(7)
        """
        return self._get_info_by_id(id, 'agent.info')

    def get_status(self, id: int) -> AgentApi.Status:
        """
        Status object containing status of Agent: waiting or running.

        :param id: Agent ID in Supervisely.
        :type id: int
        :return: Agent Status
        :rtype: :class:`Status<supervisely.api.agent_api.AgentApi.Status>`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            agent = api.agent.get_status(7)
        """
        status_str = self.get_info_by_id(id).status
        return self.Status(status_str)

    def raise_for_status(self, status):
        """
        """
        pass
